
import os, sys, pickle

SPM_LIMIT = 0x10000

class Function:
    def __init__(self, name):
        self.name = name
        self.base = 0
        self.code_size = 0
        self.stack_size = 0
        self.heap_blocks = 0
        self.root = False
        self.tracked = False
        self.init = False
        self.lib = False
        self.subtask = []

def calc_blocks(byte_size, block_size):
    return (byte_size + block_size - 1) / block_size

def main(outfile):

    num_blocks = 0
    block_size = 1
    by_name = dict()
    by_address = dict()

    def newfunc(name, base, size):
        fn = None
        if base != 0:
            fn = by_address.get(base, None)
        if fn == None:
            fn = by_name.get(name, None)
        if fn == None:
            fn = Function(name)

        fn.base = base
        fn.code_size = size
        by_name[name] = fn
        by_address[base] = fn
        return fn

    # Find root functions
    tasks = []
    for line in file("tasks.lst"):
        fields = line.split()
        if len(fields) != 0:
            assert len(fields) == 3
            name = fields[0]
            fn = newfunc(name, 0, 0)
            fn.root = True
            fn.tracked = True
            tasks.append(fields)

    for line in file("init.lst"):
        init = line.strip()
        if len(init) != 0:
            fn = newfunc(init, 0, 0)
            fn.init = True

    for line in file("lib.lst"):
        fields = line.split()
        if len(fields) != 0:
            assert len(fields) == 2
            (lib, root) = fields
            fn = newfunc(lib, 0, 0)
            fn.lib = True
            newfunc(root, 0, 0).subtask.append(fn)

    # Find functions in prog.elf that might be called as tasks or
    # called from within tasks
    try:
        fd = os.popen("mb-nm -aS capt.elf")
    except:
        fd = []

    for line in fd:
        fields = line.split()
        if len(fields) == 3:
            lbl = fields[-1]
            if lbl == "the_block_size_is":
                block_size = int(fields[0], 16)
                print 'block size %d' % block_size
            elif lbl == "the_num_blocks_is":
                num_blocks = int(fields[0], 16)
                print 'num blocks %d' % num_blocks

        if len(fields) != 4:
            continue

        if fields[2] != "T":
            continue

        base = int(fields[0], 16)
        size = int(fields[1], 16)
        name = fields[3]

        if base < SPM_LIMIT:
            continue
      
        newfunc(name, base, size)
        print 'function %s size %d' % (name, size)


    # Collect runtime information about heap and stack usage
    try:
        fd = file("capt.trc", "rt")
    except:
        fd = []

    call_stack = []
    r1_stack = []

    for line in fd:
        if line.startswith("CALL "):
            fields = line.split() 
            a = int(fields[3], 16)
            assert by_address.has_key(a), hex(a)
            fn = by_address[a]
            r1_stack.append(int(fields[5], 16))
            call_stack.append(fn)

            if len(call_stack) == 1:
                fn.root = True

        elif line.startswith("RETURN "):
            fields = line.split() 
            heap_blocks = int(fields[3], 16)
            assert len(call_stack) != 0
            fn = call_stack[-1]
            fn.heap_blocks = max(fn.heap_blocks, heap_blocks)
            fn.tracked = True

            call_stack.pop()
            r1_stack.pop()

        elif (len(call_stack) != 0) and (line.find(": r1 = ") > 0):
            assert len(r1_stack) == len(call_stack)
            fields = line.split()
            r1 = int(fields[-1], 16)

            fn = call_stack[-1]
            stack_size = r1_stack[-1] - r1
            fn.stack_size = max(fn.stack_size, stack_size)


    # Create tables
    # Dump information for every function 
    biw_h = file("biw.h", "wt")
    main_out = file(outfile, "wt")
    main_out.write("#include \"rtos.h\"\n")
    main_out.write("#include \"biw.h\"\n")

    for (name, fn) in sorted(by_name.items()):
        K = "%-15s %s"
        print K % ("name", name)
        print K % ("root?", fn.root)
        print K % ("address", hex(fn.base))
        print K % ("code size", fn.code_size)
        print K % ("stack size", fn.stack_size)

        fn.stack_blocks = sb = calc_blocks(fn.stack_size, block_size)
        fn.code_blocks = cb = calc_blocks(fn.code_size, block_size)
        if sb != 0:
            assert block_size > 1

        print K % ("heap blocks", fn.heap_blocks)
        print K % ("stack blocks", sb)
        print K % ("code blocks", cb)

        p = sb + cb + fn.heap_blocks
        assert 0 <= p <= 0xff
        assert p <= num_blocks 

        biw = p | (sb << 8) | (cb << 16) | (0x7 << 28)
        print K % ("biw code", hex(biw))
        print ''

        biw_h.write("extern const void * _BIW_%s;\n" % fn.name)
        biw_h.write("#define BIW_%s ((unsigned) (&_BIW_%s))\n" % (
                            fn.name, fn.name))
        main_out.write(r'asm(".global _BIW_%s\n"' % fn.name)
        main_out.write("\n")
        main_out.write(r'"_BIW_%s = 0x%x\n");' % (fn.name, biw))
        main_out.write("\n")

    pickle.dump(by_name, file("biw.dat", "wb"))

    biw_h.write("\n") 
    biw_h.close()


    for (name, fn) in sorted(by_name.items()):
        if fn.init or fn.root:
            main_out.write("extern void %s(void);\n" % (fn.name))

    main_out.write(r"""
int main(void)
{
    print("======================\n");
    print("main() running\n");
    print("======================\n");
""")
    for (name, fn) in sorted(by_name.items()):
        if fn.init:
            main_out.write("    %s();\n" % fn.name)
        
    main_out.write(r"""
    print("initialised\n");
""")

    for fields in tasks:
        assert len(fields) == 3
        name = fields[0]
        offset = int(fields[1])
        period = int(fields[2])
        main_out.write("    register_task(\"%s\", %s, BIW_%s, "
                "%u / TIMER_QUANTUM, %u / TIMER_QUANTUM);\n" % (
                name, name, name, offset, period))
        
    main_out.write(r"""
    print("start\n");
    kernel_start();
    return 0;
}
""")
    main_out.close()

if __name__ == "__main__":
    main(sys.argv[1])

